# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Optional

from pydantic import BaseModel, Field

from buildgrid_metering.models.dataclasses import Usage


class GetThrottlingResponse(BaseModel):
    """Summary if an identity should be throttled for the next remote execution

    Args:
        throttled (bool): if an identity should be throttled for the next remote execution
        tracked_usage (Usage): A subset of resource usage that is over the threshold
        tracked_time_window_secs: If set, this is the time window in seconds
            when the identity over consumed resource usage
    """

    throttled: bool
    tracked_usage: Optional[Usage] = None
    tracked_time_window_secs: Optional[int] = None


class PutUsageRequest(BaseModel):
    """Put the resource usage of an identity

    Args:
        operation_name (str): REAPI operation name
        usage (Usage): resource usage
    """

    operation_name: str = Field(..., description="REAPI operation name")
    usage: Usage = Field(..., description="Resource usage associated with the operation")



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import Awaitable, Callable, Dict, Optional, TypeVar
from urllib.parse import quote_plus

import aiohttp
import tenacity

from buildgrid_metering.client.auth import AuthTokenLoader
from buildgrid_metering.client.exceptions import MeteringServiceClientError, MeteringServiceHTTPError
from buildgrid_metering.client.retry import RetryConfig
from buildgrid_metering.models.api import GetThrottlingResponse, PutUsageRequest
from buildgrid_metering.models.dataclasses import Identity, Usage

T = TypeVar("T")

USAGE_API_URL_PREFIX = "/v1/usage/"
THROTTLING_API_URL_PREFIX = "/v1/throttling/"
IDENTITY_EMPTY_INSTANCE_URL_TEMPLATE = "{workflow}/{actor}/{subject}"
IDENTITY_URL_TEMPLATE = "{instance}/" + IDENTITY_EMPTY_INSTANCE_URL_TEMPLATE


class MeteringServiceClient:
    """Python client of BuildGrid Metering Service"""

    _base_url: str
    _token_loader: AuthTokenLoader
    _request_timeout: aiohttp.ClientTimeout
    _headers: Dict[str, str]
    _retry_config: RetryConfig

    def __init__(
        self,
        base_url: str,
        token_loader: AuthTokenLoader,
        request_timeout: float = 5.0,
        headers: Optional[Dict[str, str]] = None,
        retry_config: RetryConfig = RetryConfig(),
    ) -> None:
        """Initialize the client

        Args:
            base_url (str): base URL of the service,
                e.g. http://localhost:8000 or https://somedomain.com
            token_loader (AuthTokenLoader): token loader
            request_timeout (float, optional): HTTP Request timeout . Defaults to 5.0.
            headers (Optional[dict[str, str]], optional): default headers.
                Defaults to None.
            retry_configs: See `buildgrid_metering.client.retry.RetryConfig`
                for details and default values.
        """
        self._base_url = base_url
        self._token_loader = token_loader
        self._request_timeout = aiohttp.ClientTimeout(total=request_timeout)
        self._headers = headers or {}
        self._retry_config = retry_config

    async def put_usage(self, identity: Identity, operation_name: str, usage: Usage) -> None:
        url = self._create_url(USAGE_API_URL_PREFIX, identity)
        request = PutUsageRequest(operation_name=operation_name, usage=usage)

        async def request_func(session: aiohttp.ClientSession) -> None:
            async with session.put(
                url,
                json=request.dict(exclude_defaults=True),
            ) as response:
                if response.status != 204:
                    raise MeteringServiceHTTPError(response.status, await response.text())

        return await self._make_request(request_func)

    async def get_throttling(self, identity: Identity) -> GetThrottlingResponse:
        url = self._create_url(
            THROTTLING_API_URL_PREFIX,
            identity,
        )

        async def request_func(session: aiohttp.ClientSession) -> GetThrottlingResponse:
            async with session.get(url) as response:
                if response.status != 200:
                    raise MeteringServiceHTTPError(response.status, await response.text())
                return GetThrottlingResponse(**(await response.json()))

        return await self._make_request(request_func)

    async def _make_request(
        self,
        request_func: Callable[[aiohttp.ClientSession], Awaitable[T]],
    ) -> T:
        headers = dict(self._headers)
        auth_token = self._token_loader.get_token()
        if auth_token is not None:
            headers["Authorization"] = "Bearer " + auth_token

        def should_retry_on_http_error(exc: BaseException) -> bool:
            if not isinstance(exc, MeteringServiceHTTPError):
                return False
            return exc.http_status in self._retry_config.http_statuses

        @tenacity.retry(
            stop=tenacity.stop_after_attempt(self._retry_config.max_attempts),
            retry=tenacity.retry_if_exception(should_retry_on_http_error)
            | tenacity.retry_if_exception_type(self._retry_config.exception_types)
            | tenacity.retry_if_exception_cause_type(self._retry_config.cause_exception_types),
            wait=tenacity.wait_exponential(
                multiplier=self._retry_config.multiplier,
                exp_base=self._retry_config.exp_base,
                max=self._retry_config.max_wait,
            ),
            reraise=True,
        )
        async def _request() -> T:
            try:
                async with aiohttp.ClientSession(
                    self._base_url, headers=headers, timeout=self._request_timeout
                ) as session:
                    return await request_func(session)
            except (aiohttp.ClientError, asyncio.TimeoutError) as e:
                raise MeteringServiceClientError(repr(e)) from e

        return await _request()

    def _create_url(self, url_prefix: str, identity: Identity) -> str:
        if identity.instance == "":
            return url_prefix + IDENTITY_EMPTY_INSTANCE_URL_TEMPLATE.format(
                workflow=quote_plus(identity.workflow),
                actor=quote_plus(identity.actor),
                subject=quote_plus(identity.subject),
            )
        return url_prefix + IDENTITY_URL_TEMPLATE.format(
            instance=quote_plus(identity.instance),
            workflow=quote_plus(identity.workflow),
            actor=quote_plus(identity.actor),
            subject=quote_plus(identity.subject),
        )



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from enum import Enum
from typing import Optional

from cachetools.func import ttl_cache

from buildgrid_metering.client.exceptions import MeteringServiceClientError


class AuthTokenMode(str, Enum):
    CONSTANT = "constant"
    FILEPATH = "filepath"
    NONE = "none"


@dataclass
class AuthTokenConfig:
    """Config on how to load the JWT token.
    if `mode == CONSTANT`, the token will always be the `value`
    if `mode == FILEPATH`, it reads the file located at `value`
    if `mode == NONE`, it returns `None`
    """

    mode: AuthTokenMode
    value: str


class AuthTokenLoader:
    """A class that loads JWT authentication token based on the config"""

    _config: AuthTokenConfig

    def __init__(self, config: AuthTokenConfig) -> None:
        self._config = config

    def get_token(self) -> Optional[str]:
        """Get the JWT authentication token

        Raises:
            ValueError: if the `mode` is not supported

        Returns:
            str | None: JWT if `mode` is not NONE
        """
        if self._config.mode == AuthTokenMode.NONE:
            return None
        if self._config.mode == AuthTokenMode.CONSTANT:
            return self._config.value
        if self._config.mode == AuthTokenMode.FILEPATH:
            return _load_token_from_file(self._config.value)
        raise ValueError(f"Unsupported AuthTokenMode: {self._config.mode}")


@ttl_cache(maxsize=1, ttl=60 * 60)
def _load_token_from_file(path: str) -> str:
    try:
        with open(path, mode="r") as token_file:
            return token_file.read().strip()
    except Exception as e:
        raise MeteringServiceClientError(f"Cannot read token from filepath: {path}") from e



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MeteringServiceError(RuntimeError):
    """Base class of all errors"""


class MeteringServiceHTTPError(MeteringServiceError):
    """Errors indicated by the HTTP status, e.g., 400, 503"""

    http_status: int
    message: str

    def __init__(self, http_status: int, message: str) -> None:
        super().__init__(http_status, message)
        self.http_status = http_status
        self.message = message


class MeteringServiceClientError(MeteringServiceError):
    """Client connection errors, e.g. timeout, connection closed"""

    message: str

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message



from .auth import AuthTokenConfig, AuthTokenLoader, AuthTokenMode
from .client import MeteringServiceClient
from .retry import RetryConfig

__all__ = [
    "MeteringServiceClient",
    "AuthTokenConfig",
    "AuthTokenLoader",
    "AuthTokenMode",
    "RetryConfig",
]



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import pytest

from buildgrid_metering.client.auth import AuthTokenConfig, AuthTokenLoader, AuthTokenMode
from buildgrid_metering.client.exceptions import MeteringServiceClientError


@pytest.mark.asyncio
async def test_auth_loader_constant():
    # GIVEN
    config = AuthTokenConfig(AuthTokenMode.CONSTANT, "sometoken")
    loader = AuthTokenLoader(config)

    # WHEN
    token = loader.get_token()

    # THEN
    assert token == "sometoken"


@pytest.mark.asyncio
async def test_auth_loader_none():
    # GIVEN
    config = AuthTokenConfig(AuthTokenMode.NONE, "")
    loader = AuthTokenLoader(config)

    # WHEN
    token = loader.get_token()

    # THEN
    assert token is None


@pytest.mark.asyncio
@mock.patch("buildgrid_metering.client.auth.open")
async def test_auth_loader_filepath(mock_open_file):
    # GIVEN
    mock_token_file = mock.MagicMock()
    mock_token_file.read.return_value = "sometoken"
    mock_open_file.return_value.__enter__.return_value = mock_token_file

    config = AuthTokenConfig(AuthTokenMode.FILEPATH, "/tmp/token")
    loader = AuthTokenLoader(config)

    # WHEN
    token = loader.get_token()

    # THEN
    assert token == "sometoken"


@pytest.mark.asyncio
@mock.patch("buildgrid_metering.client.auth.open")
async def test_auth_loader_filepath_not_readable(mock_open_file):
    # GIVEN
    mock_open_file.return_value.__enter__.side_effect = OSError("cannot open file")

    config = AuthTokenConfig(AuthTokenMode.FILEPATH, "/tmp/nottoken")
    loader = AuthTokenLoader(config)

    # WHEN
    with pytest.raises(MeteringServiceClientError):
        await loader.get_token()



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import dataclass
from typing import Tuple, Type


@dataclass(frozen=True)
class RetryConfig:
    """RetryConfig for MeteringServiceClient"""

    max_attempts: int = 0
    """Max attempts to retry the API call. Default to 0 (no retry)."""
    exp_base: float = 1.5
    """Exponential base before retry. Default to 1.5"""
    multiplier: float = 1.0
    """Multiplier for exponential back-off. Defaults to 1.0"""
    max_wait: float = 30.0
    """Maximum of wait time in seconds between retries. Default to 30.0"""
    http_statuses: Tuple[int, ...] = ()
    """HTTP status codes to retry, e.g. 503. Default to empty."""
    exception_types: Tuple[Type[Exception], ...] = ()
    """Exception types to retry. Default to empty."""
    cause_exception_types: Tuple[Type[Exception], ...] = ()
    """Cause exception types to retry. Default to empty."""



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import abc
from typing import Dict, TypeVar, Union

from pydantic import BaseModel


class Identity(BaseModel):
    """Identity that consumes build resource

    Args:
        Instance: REAPI service instance, e.g., "dev"
        Workflow: The associated workflow of the operation, e.g., "build"
        Actor: The tooling or agent that submitted the operation, e.g., "my-build-tool"
        Subject: The end user that submitted the operation, e.g. "user1"
    """

    instance: str
    workflow: str
    actor: str
    subject: str


R = TypeVar("R", bound="ResourceUsage")


class ResourceUsage(BaseModel, abc.ABC):
    @abc.abstractmethod
    def combine(self: R, other: R) -> R:
        """Combine two resource usages of the same type"""

    @abc.abstractmethod
    def get_throttled(self: R, threshold: R) -> R:
        """Get a subset of this resource type that exceeds the predefined threshold
        while attributes that are <= threshold are default-valued."""

    def __add__(self: R, other: R) -> R:
        return self.combine(other)


N = TypeVar("N", bound="NumericResourceUsage")


class NumericResourceUsage(ResourceUsage):
    """A subtype of ResourceUsage where each attribute is either int or float.
    It provides reasonable default methods.
    """

    def combine(self: N, other: N) -> N:
        self_dict = self.dict(exclude_unset=True)
        other_dict = other.dict(exclude_unset=True)

        combined_dict: Dict[str, Union[int, float]] = {
            attr: self_dict.get(attr, 0) + other_dict.get(attr, 0) for attr in self_dict.keys() | other_dict.keys()
        }

        return type(self)(**combined_dict)

    def get_throttled(self: N, threshold: N) -> N:
        threshold_dict = threshold.dict(exclude_unset=True)
        self_dict = self.dict()

        throttled = {k: self_dict[k] for k, v in threshold_dict.items() if self_dict[k] > v}

        return type(self)(**throttled)


class ComputingUsage(NumericResourceUsage):
    """Basic computing resource usage defined in getrusage(2)"""

    utime: int = 0
    stime: int = 0
    maxrss: int = 0
    inblock: int = 0
    oublock: int = 0


class CASUsage(NumericResourceUsage):
    """IO usage of ContentAddressableStorage"""

    read_bytes: int = 0
    write_bytes: int = 0


class Usage(ResourceUsage):
    """
    The aggregation of all kinds of usages
    """

    computing: ComputingUsage = ComputingUsage()
    cas: CASUsage = CASUsage()

    def combine(self, other: Usage) -> Usage:
        computing = self.computing + other.computing
        cas = self.cas + other.cas

        return Usage(computing=computing, cas=cas)

    def get_throttled(self: Usage, threshold: Usage) -> Usage:
        computing = self.computing.get_throttled(threshold.computing)
        cas = self.cas.get_throttled(threshold.cas)

        return Usage(computing=computing, cas=cas)

    def __str__(self) -> str:
        return f"computing: [ {self.computing} ], cas: [ {self.cas} ]"



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from unittest import mock

import aiohttp
import pytest

from buildgrid_metering.client.auth import AuthTokenConfig, AuthTokenLoader, AuthTokenMode
from buildgrid_metering.client.client import MeteringServiceClient
from buildgrid_metering.client.exceptions import MeteringServiceClientError, MeteringServiceHTTPError
from buildgrid_metering.client.retry import RetryConfig
from buildgrid_metering.models.api import GetThrottlingResponse
from buildgrid_metering.models.dataclasses import ComputingUsage, Identity, Usage

MOCK_IDENTITY = Identity(instance="dev", workflow="build", actor="tool", subject="username")
MOCK_USAGE = Usage(computing=ComputingUsage(utime=123))


@pytest.fixture
def mock_http_session():
    with mock.patch("buildgrid_metering.client.client.aiohttp.ClientSession") as session_cls:
        session = mock.MagicMock()
        session_cls.return_value.__aenter__.return_value = session
        yield session


@pytest.fixture
def client():
    token_loader = AuthTokenLoader(AuthTokenConfig(AuthTokenMode.CONSTANT, "token"))
    yield MeteringServiceClient("http://localhost:8000", token_loader)


@pytest.mark.asyncio
async def test_put_usage(mock_http_session, client: MeteringServiceClient):
    # GIVEN
    mock_response = mock.AsyncMock
    mock_response.status = 204
    mock_http_session.put.return_value.__aenter__.return_value = mock_response

    # WHEN
    await client.put_usage(MOCK_IDENTITY, "op", MOCK_USAGE)

    # THEN
    # no exception is raised


@pytest.mark.asyncio
async def test_put_usage_unprocessable_content(mock_http_session, client: MeteringServiceClient):
    # GIVEN
    mock_response = mock.AsyncMock()
    mock_response.status = 422
    mock_response.text.return_value = "request cannot be validated"
    mock_http_session.put.return_value.__aenter__.return_value = mock_response

    # WHEN
    with pytest.raises(MeteringServiceHTTPError) as e:
        await client.put_usage(MOCK_IDENTITY, "op", MOCK_USAGE)

    # THEN
    assert e.value.http_status == 422


@pytest.mark.asyncio
async def test_put_usage_timeout(mock_http_session, client: MeteringServiceClient):
    # GIVEN
    mock_http_session.put.return_value.__aenter__.side_effect = aiohttp.ServerTimeoutError()

    # WHEN
    with pytest.raises(MeteringServiceClientError):
        await client.put_usage(MOCK_IDENTITY, "op", MOCK_USAGE)


@pytest.mark.asyncio
async def test_get_throttling(mock_http_session, client: MeteringServiceClient):
    # GIVEN
    mock_response = GetThrottlingResponse(throttled=False)
    mock_http_response = mock.AsyncMock()
    mock_http_response.status = 200
    mock_http_response.json.return_value = mock_response.dict(exclude_unset=True)
    mock_http_session.get.return_value.__aenter__.return_value = mock_http_response

    # WHEN
    response = await client.get_throttling(MOCK_IDENTITY)

    # THEN
    assert response == mock_response


@pytest.mark.asyncio
async def test_get_throttling_unavailable(mock_http_session, client: MeteringServiceClient):
    # GIVEN
    mock_http_response = mock.AsyncMock()
    mock_http_response.status = 503
    mock_http_response.text.return_value = "service unavailable"
    mock_http_session.get.return_value.__aenter__.return_value = mock_http_response

    # WHEN
    with pytest.raises(MeteringServiceHTTPError) as e:
        await client.get_throttling(MOCK_IDENTITY)

    # THEN
    assert e.value.http_status == 503


@pytest.mark.asyncio
async def test_get_throttling_retry_failure(mock_http_session, client: MeteringServiceClient):
    # GIVEN
    client._retry_config = RetryConfig(
        max_attempts=3,
        multiplier=0.0,
        http_statuses=(503,),
    )
    mock_http_response = mock.AsyncMock()
    mock_http_response.status = 503
    mock_http_response.text.return_value = "service unavailable"
    mock_http_session.get.return_value.__aenter__.return_value = mock_http_response

    # WHEN
    with pytest.raises(MeteringServiceHTTPError) as e:
        await client.get_throttling(MOCK_IDENTITY)

    # THEN
    assert e.value.http_status == 503
    assert mock_http_session.get.call_count == 3


@pytest.mark.asyncio
async def test_get_throttling_retry_success(mock_http_session, client: MeteringServiceClient):
    # GIVEN
    client._retry_config = RetryConfig(
        max_attempts=3,
        multiplier=0.0,
        http_statuses=(503,),
    )
    mock_503_response = mock.AsyncMock()
    mock_503_response.status = 503
    mock_503_response.text.return_value = "service unavailable"

    mock_200_response = mock.AsyncMock()
    mock_200_response.status = 200
    mock_200_response.json.return_value = GetThrottlingResponse(throttled=False).dict(exclude_unset=True)
    mock_http_session.get.return_value.__aenter__.side_effect = [
        mock_503_response,
        mock_200_response,
    ]

    # WHEN
    response = await client.get_throttling(MOCK_IDENTITY)

    # THEN
    # no exception
    assert not response.throttled
    assert mock_http_session.get.call_count == 2


@pytest.mark.asyncio
async def test_get_throttling_retry_on_cause_exception(mock_http_session, client: MeteringServiceClient):
    # GIVEN
    client._retry_config = RetryConfig(
        max_attempts=3,
        multiplier=0.0,
        cause_exception_types=(asyncio.TimeoutError,),
    )

    def _raise():
        raise MeteringServiceClientError("timeout") from asyncio.TimeoutError()

    mock_http_session.get.return_value.__aenter__.side_effect = _raise

    # WHEN
    with pytest.raises(MeteringServiceClientError):
        await client.get_throttling(MOCK_IDENTITY)

    # THEN
    assert mock_http_session.get.call_count == 3


@pytest.mark.asyncio
async def test_get_throttling_retry_on_exception(mock_http_session, client: MeteringServiceClient):
    # GIVEN
    client._retry_config = RetryConfig(
        max_attempts=3,
        multiplier=0.0,
        exception_types=(MeteringServiceClientError,),
    )

    mock_http_session.get.return_value.__aenter__.side_effect = aiohttp.ServerTimeoutError()

    # WHEN
    with pytest.raises(MeteringServiceClientError):
        await client.get_throttling(MOCK_IDENTITY)

    # THEN
    assert mock_http_session.get.call_count == 3
